<?php

namespace Netzmacher\Launch\Controller;

use TYPO3\CMS\Core\Resource\ResourceFactory;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\PathUtility;
use TYPO3\CMS\Extbase\Mvc\Controller\ActionController;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2017 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Class for frontend editing
 *
 * @package TYPO3
 * @subpackage launch
 * @author Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @version 2.0.0
 * @since 2.0.0
 */
class AbstractController extends ActionController
{

	/**
	 * @var string Extension key
	 */
	public $extKey = 'launch';

	/**
	 * @var array Extension Configuration from ext_conf_template.txt
	 */
	private $_aExtConf;

	/**
	 * @var array 
	 */
	private $_aPrompts;

	/**
	 * @var array 
	 */
	private $_aSearchAndReplace;

	/**
	 * @var int Last inserted id
	 */
	private $_aTableSQLLastID;

	/**
	 * @var array with table.record = uid
	 */
	private $_aTableRecordUid;

	/**
	 * @var integer unix timestamp
	 */
	private $_iCurrTime;

	/**
	 * @var integer current pid
	 */
	private $_iCurrPid;

	/**
	 * @var integer current uid
	 */
	private $_iCurrUid;

	/**
	 * @var integer import id (yymmddhhss)
	 */
	private $_iLaunchId;

	/**
	 * @var TYPO3\CMS\Frontend\ContentObject\ContentObjectRenderer
	 */
	private $_oCObj;

	/**
	 * @var TYPO3\CMS\Core\Imaging\GraphicalFunctions
	 */
	private $_oGraphFuncObj;

	/**
	 * @var string current table
	 */
	private $_sCurrTable;

	/**
	 * @var string path to the lanugage file
	 */
	private $_sLanguageFilePath = 'Resources/Private/Language/Installer.xlf';

	/**
	 * _SQLError( ) :
	 *
	 * @param string $query
	 * @param integer $line
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _SQLError( $query, $line = __LINE__ )
	{
		$sqlError = $GLOBALS[ 'TYPO3_DB' ]->sql_error();
		if( empty( $sqlError ) )
		{
			return;
		}

		$sqlErrno = $GLOBALS[ 'TYPO3_DB' ]->sql_errno();
		$header = 'SQL-ERROR';
		$text = ''
						. '<p>'
						. '  Error: ' . $sqlError . '<br />'
						. '  Error number: ' . $sqlErrno . '<br />'
						. '  Query: ' . $query . '<br />'
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, __METHOD__, $line );
	}

	/**
	 * _SQLDelete( ) :
	 *
	 * @param string $table
	 * @param string $where
	 * @param integer $line
	 * @param boolean $bPromptQueryOnly
	 * @return void
	 * @access private
	 * @version 8.0.18
	 * @since 8.0.18
	 */
	private function _SQLDelete( $table, $where, $line, $bPromptQueryOnly = 0 )
	{
		$query = $GLOBALS[ 'TYPO3_DB' ]->DELETEquery(
						$table
						, $where
		);

		switch( true )
		{
			case($bPromptQueryOnly == 1):
				var_dump( __METHOD__, __LINE__, $query );
				break;
			case(FALSE):
			default:
				$GLOBALS[ 'TYPO3_DB' ]->exec_DELETEquery(
								$table
								, $where
				);
				$this->_SQLError( $query, $line );
				break;
		}
	}

	/**
	 * _SQLInsert( ) :
	 *
	 * @param string $table
	 * @param integer $data
	 * @param integer $line
	 * @param boolean $bPromptQueryOnly
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _SQLInsert( $table, $data, $line, $bPromptQueryOnly = 0 )
	{
		$query = $GLOBALS[ 'TYPO3_DB' ]->INSERTquery(
						$table
						, $data
		);
		//var_dump( __METHOD__, $line, $query, $data );
		switch( $bPromptQueryOnly )
		{
			case(TRUE):
//				$this->_DRSprompt( 'Query: ' . $query, $line );
//				$this->_DRSprompt( 'Query is for prompt only and will not executed.', $line );
				var_dump( __METHOD__, $line, $query );
				break;
			case(FALSE):
			default:
//				$this->_DRSprompt( 'Query: ' . $query, $line );
				$GLOBALS[ 'TYPO3_DB' ]->exec_INSERTquery(
								$table
								, $data
				);
				$this->_SQLError( $query, $line );
				$this->_aTableSQLLastID[ $table ] = $GLOBALS[ 'TYPO3_DB' ]->sql_insert_id();
				break;
		}
	}

	/**
	 * _SQLSelect( ) :
	 *
	 * @param string $table
	 * @param string $where
	 * @param integer $line
	 * @param boolean $bPromptQueryOnly
	 * @return void
	 * @access private
	 * @version 8.0.18
	 * @since 8.0.18
	 */
	private function _SQLSelect( $fields, $table, $where, $groupBy = '', $orderBy = '', $limit = '', $line = __LINE__, $bPromptQueryOnly = 0 )
	{
		$query = $GLOBALS[ 'TYPO3_DB' ]->SELECTquery(
						$fields
						, $table
						, $where
						, $groupBy = ''
						, $orderBy = ''
						, $limit = ''
		);

		switch( true )
		{
			case($bPromptQueryOnly == 1):
				var_dump( __METHOD__, $line, $query );
				return;
			case(FALSE):
			default:
				$res = $GLOBALS[ 'TYPO3_DB' ]->exec_SELECTquery(
								$fields
								, $table
								, $where
								, $groupBy = ''
								, $orderBy = ''
								, $limit = ''
				);
				$this->_SQLError( $query, $line );
				return $res;
		}
	}

	/**
	 * _SQLUpdate( ) :
	 *
	 * @param string $table
	 * @param integer $uid
	 * @param integer $data
	 * @param integer $line
	 * @param boolean $bPromptQueryOnly
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _SQLUpdate( $table, $where, $data, $uid, $line, $bPromptQueryOnly = 0 )
	{
		$query = $GLOBALS[ 'TYPO3_DB' ]->UPDATEquery(
						$table
						, $where
						, $data
		);

		switch( $bPromptQueryOnly )
		{
			case(TRUE):
				var_dump( __METHOD__, $line, $query );
				break;
			case(FALSE):
			default:
				$GLOBALS[ 'TYPO3_DB' ]->exec_UPDATEquery(
								$table
								, $where
								, $data
				);
				$this->_aTableSQLLastID[ $table ] = $uid;
				$this->_SQLError( $query, $line );
				break;
		}
	}

	/**
	 * _cObjGetSingleArray( ) :
	 *
	 * @param array $srceRecords
	 * @return array $destRecords
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _cObjGetSingleArray( $srceRecords )
	{
		foreach( ( array ) $srceRecords as $key => $record )
		{
			switch( true )
			{
				case(is_array( $record )):
					$destRecords[ $key ] = $this->_cObjGetSingle( $record );
					break;
				case(!is_array( $record )):
				default:
					$destRecords[ $key ] = $record;
					break;
			}
		}

		return $destRecords;
	}

	/**
	 * _cObjGetSingle( ) :
	 *
	 * @param array $record
	 * @return string $value
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _cObjGetSingle( $record )
	{
		$name = $record[ '_typoScriptNodeValue' ];
		$conf = $this->_oTypoScriptService->convertPlainArrayToTypoScriptArray( $record );
		$value = $this->_oCObj->cObjGetSingle( $name, $conf );

		return $value;
	}

	/**
	 * _clearCache( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _clearCache()
	{
		$query = 'TRUNCATE `cf_cache_pages`';
		$GLOBALS[ 'TYPO3_DB' ]->sql_query( $query );
		$this->_SQLError( $query, __LINE__ );
	}

	/**
	 * _deleteDirectory( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _deleteDirectory( $destPath )
	{
		$files = array_diff( scandir( $destPath ), array( '..', '.' ) );
		$launchId = '-LaunchId' . $this->_iLaunchId . '-';

		foreach( $files as $file )
		{
			$pos = strpos( $file, $launchId );
			if( $pos === false )
			{
				continue;
			}
			unlink( $destPath . '/' . $file );
		}
	}

	/**
	 * _deleteDirectories( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _deleteDirectories()
	{
		$directories = $this->settings[ 'configuration' ][ 'directories' ];
		unset( $directories[ '_typoScriptNodeValue' ] );

		foreach( $directories as $properties )
		{
			$destPath = $properties[ 'dest' ];
			$this->_deleteDirectory( $destPath );
		}
	}

	/**
	 * _deleteTable( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _deleteTable( $table )
	{
		$where = "tx_launch_uid LIKE '%" . $this->_iLaunchId . "%'";
		$bPromptQueryOnly = 0;

		$this->_SQLDelete( $table, $where, __LINE__, $bPromptQueryOnly );
	}

	/**
	 * _deleteTables( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _deleteTables()
	{
		$tables = $this->settings[ 'configuration' ][ 'tables' ];
		unset( $tables[ '_typoScriptNodeValue' ] );

		foreach( ( array ) array_keys( $tables ) as $table )
		{
			$this->_deleteTable( $table );
		}
	}

	/**
	 * _directories( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _directories()
	{
		$directories = $this->settings[ 'configuration' ][ 'directories' ];
		unset( $directories[ '_typoScriptNodeValue' ] );

		foreach( $directories as $directory )
		{
			$this->_directory( $directory );
		}
	}

	/**
	 * _directory( ) :
	 *
	 * @param string $directory
	 * @param array $properties
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _directory( $directory )
	{
		$dest = $directory[ 'dest' ];
		$srce = $directory[ 'srce' ];
		$this->_directoryCopy( $srce, $dest );
	}

	/**
	 * _directoryCopy( ) :
	 *
	 * @return	void
	 * @access private
	 * @version 8.0.0
	 * @since 1.0.0
	 */
	private function _directoryCopy( $srceDir, $destDir, $relative = 1 )
	{
		if( $relative )
		{
			$sPathDest = PATH_site . $destDir;
			$sPathSrce = PATH_site . $srceDir;
		}
		else
		{
			$sPathDest = $destDir;
			$sPathSrce = $srceDir;
		}
		// General values

		$this->_directoryMkDir( $sPathDest );

		$files = array_diff( scandir( $sPathSrce ), array( '..', '.' ) );
		foreach( $files as $file )
		{
			$srceFile = $sPathSrce . '/' . $file;
			$this->_directoryCopyFile( $srceFile, $sPathDest );
		}
	}

	/**
	 * _directoryCopyFile( ) :
	 *
	 * @return	void
	 * @access private
	 * @version 8.0.0
	 * @since 1.0.0
	 */
	private function _directoryCopyFile( $srceFile, $destPath )
	{
		$destFile = $this->_fileToDest( $srceFile, $destPath );

		$bSuccess = copy( $srceFile, $destFile );
		if( $bSuccess )
		{
			return;
		}

		$header = 'FATAL ERROR';
		$text = ''
						. '<p>'
						. '  Can\'t copy file.<br />'
						. '  Source: "' . $srceFile . '".<br />'
						. '  Destination: "' . $destFile . '".<br />'
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
	}

	/**
	 * _fileToDest( ) :
	 *
	 * @return	string
	 * @access private
	 * @version 8.0.0
	 * @since 1.0.0
	 */
	private function _fileToDest( $srceFile, $destPath = null )
	{
//		var_dump(__METHOD__, __LINE__, $srceFile, $destPath);
		$aFile = pathinfo( $srceFile );
		$extension = $aFile[ 'extension' ];
		$filename = $aFile[ 'filename' ];

		$destFilename = ''
						. $filename
						. '-LaunchId' . $this->_iLaunchId
						. '-' . $this->_iCurrTime
						. '.' . $extension
		;

		if( empty( $destPath ) )
		{
			return $destFilename;
		}

		$destFile = $destPath . '/' . $destFilename;

		return $destFile;
	}

	/**
	 * _directoryMkDir( ) :
	 *
	 * @param string $dir
	 * @return	void
	 * @access private
	 * @version 8.0.0
	 * @since 1.0.0
	 */
	private function _directoryMkDir( $dir )
	{
		if( is_dir( $dir ) )
		{
			return;
		}

		$oldmask = umask( 0 );
		if( mkdir( $dir, 02770, true ) )
		{
			umask( $oldmask );
			return;
		}

		umask( $oldmask );
		$header = 'FATAL ERROR: Can\'t create directory!';
		$text = ''
						. '<p>'
						. '  Failed to create "' . $dir . '"' . PHP_EOL
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
	}

	/**
	 * _extensionRecord( ) :
	 *
	 * @param string $extension
	 * @param array $aRecord
	 * @param string $constraints
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecord( $extension, $aRecord, $constraints )
	{

		switch( $constraints )
		{
			case('conflicts'):
				$this->_extensionRecordConflicts( $extension, $aRecord );
				break;
			case('depends'):
				$this->_extensionRecordDepends( $extension, $aRecord );
				break;
			case('suggests'):
				$this->_extensionRecordSuggests( $extension, $aRecord );
				break;
			default:
				$header = 'FATAL ERROR: undefined constraints';
				$text = ''
								. '<p>'
								. '  undefined constraints: "' . $constraints . '".'
								. '</p>'
				;
				$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
				break;
		}
	}

	/**
	 * _extensionRecordConflicts( ) :
	 *
	 * @param string $extension
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecordConflicts( $extension, $aRecord )
	{
		$aRecord = $this->_cObjGetSingleArray( $aRecord );
		$aRecord = $this->_extensionRecordDefaults( $aRecord, 'conflicts' );

		if( !\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::isLoaded( $extension ) )
		{
			return;
		}
		$prompt = $aRecord[ 'title' ] . ' (' . $extension . '): ' . $aRecord[ 'prompt' ];
		$this->_aPrompts[ 'alert' ][] = $prompt;
//		var_dump( __METHOD__, __LINE__, $extension, $aRecord );
//		die();
	}

	/**
	 * _extensionRecordDepends( ) :
	 *
	 * @param string $extension
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecordDepends( $extension, $aRecord )
	{
		$aRecord = $this->_cObjGetSingleArray( $aRecord );
		$aRecord = $this->_extensionRecordDefaults( $aRecord, 'depends' );

		if( \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::isLoaded( $extension ) )
		{
			return;
		}
		$prompt = $aRecord[ 'title' ] . ' (' . $extension . '): ' . $aRecord[ 'prompt' ];
		$this->_aPrompts[ 'alert' ][] = $prompt;
//		var_dump( __METHOD__, __LINE__, $extension, $aRecord );
//		die();
	}

	/**
	 * _extensionRecordSuggests( ) :
	 *
	 * @param string $extension
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecordSuggests( $extension, $aRecord )
	{
		$aRecord = $this->_cObjGetSingleArray( $aRecord );
		$aRecord = $this->_extensionRecordDefaults( $aRecord, 'suggests' );

		if( \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::isLoaded( $extension ) )
		{
			return;
		}
		$prompt = $aRecord[ 'title' ] . ' (' . $extension . '): ' . $aRecord[ 'prompt' ];
		$this->_aPrompts[ 'info' ][] = $prompt;
//		var_dump( __METHOD__, __LINE__, $extension, $aRecord );
//		die();
	}

	/**
	 * _extensionRecordDefaults( ) :
	 *
	 * @param array $aRecord
	 * @return array $aRecord
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecordDefaults( $aRecord, $constraints )
	{
		// configuration.extensions.conflicts.prompts
		$defaults = $this->settings[ 'configuration' ][ 'extensions' ][ $constraints ][ 'defaults' ];
		$defaults = $this->_cObjGetSingleArray( $defaults );

		unset( $defaults[ '_typoScriptNodeValue' ] );

		if( !is_array( $defaults ) )
		{
			return $aRecord;
		}

		$aRecord = ( array ) $aRecord + ( array ) $defaults;
		return $aRecord;
	}

	/**
	 * _extensionRecords( ) :
	 *
	 * @param array $aRecords
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecords( $aRecords, $constraints )
	{
		foreach( $aRecords as $extension => $aRecord )
		{
			$aRecord = $this->_extensionRecord( $extension, $aRecord, $constraints );
		}
	}

	/**
	 * _extensionRecursive( ) :
	 *
	 * @param object $iterator
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensionRecursive( $iterator, $constraints )
	{
		while( $iterator->valid() )
		{
			if( !$iterator->hasChildren() )
			{
				$iterator->next();
				continue;
			}

			switch( true )
			{
				case($iterator->key() == 'records'):
					$aRecords = $iterator->current();
					$this->_extensionRecords( $aRecords, $constraints );
					break;
				default:
					// do nothing
					break;
			}

			self::_extensionRecursive( $iterator->getChildren(), $constraints );
			$iterator->next();
		}
	}

	/**
	 * _extensions( ) :
	 *
	 * @return boolean
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _extensions()
	{
		$extensions = $this->settings[ 'configuration' ][ 'extensions' ];
		unset( $extensions[ '_typoScriptNodeValue' ] );

		foreach( array_keys( ( array ) $extensions ) as $constraints )
		{
			$pathToXml = $this->settings[ 'configuration' ][ 'extensions' ][ $constraints ][ 'xml' ];
			if( empty( $pathToXml ) )
			{
				return;
			}
			$aXml = $this->_getXmlFileAsArray( $pathToXml );
			$iterator = new \RecursiveArrayIterator( $aXml );
			iterator_apply( $iterator, 'Netzmacher\Launch\Controller\AbstractController::_extensionRecursive', array( $iterator, $constraints ) );
		}
//		var_dump( __METHOD__, __LINE__ );
//		die();
	}

	/**
	 * _getUidLaunch( ) :
	 *
	 * @return integer 
	 * @access protected
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	protected function _getUidLaunch()
	{
		$fields = 'max(uid)';
		$table = 'pages';
		$where = "pid = " . $this->_iCurrUid . " AND tx_launch_uid LIKE '%" . $this->_iLaunchId . "%'";
		$groupBy = null;
		$orderBy = null;
		$limit = 1;
		$bPromptQueryOnly = 0;

		$res = $this->_SQLSelect( $fields, $table, $where, $groupBy, $orderBy, $limit, __LINE__, $bPromptQueryOnly );
		$row = $GLOBALS[ 'TYPO3_DB' ]->sql_fetch_row( $res );
		$uidLaunch = $row[ 0 ];

		return $uidLaunch;
	}

	/**
	 * _getXmlFileAsArray( ) :
	 *
	 * @param string $pathToXml
	 * @return array
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _getXmlFileAsArray( $pathToXml )
	{
		$filename = PATH_site . $pathToXml;

		$this->_zzDieWoFile( $filename, __METHOD__, __LINE__ );
		$handle = fopen( $filename, "r" );
		$sXml = fread( $handle, filesize( $filename ) );
		fclose( $handle );

		$aXml = GeneralUtility::xml2array( $sXml );

		if( is_array( $aXml ) )
		{
			return $aXml;
		}

		$header = 'FATAL ERROR: xml-file isn\'t proper';
		$text = ''
						. '<p>'
						. '  The xml-file ' . $filename . ' isn\'t proper!<br />'
						. '  Prompt "' . ( string ) $aXml . '"<br />'
						. '  Please take care of a proper xml-format.<br />'
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
	}

	/**
	 * _initExtConf( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initExtConf()
	{
		$this->_aExtConf = unserialize( $GLOBALS[ 'TYPO3_CONF_VARS' ][ 'EXT' ][ 'extConf' ][ 'browser' ] );
	}

	/**
	 * _initLL( ) : Init localisation
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initLL()
	{
		$this->_initObjectPlugin();
		$this->_oPlugin->pi_loadLL( 'EXT:' . $this->extKey . '/' . $this->_sLanguageFilePath );
	}

	/**
	 * _initObject( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObject()
	{
		$this->_initObjectGraphFuncObj();
		$this->_initObjectObjectManager();
		$this->_initObjectCObj();
		$this->_initObjectTypoScriptService();
	}

	/**
	 * _initObjectCObj( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObjectCObj()
	{
		$this->_oCObj = $this->_oObjectManager->get( 'TYPO3\\CMS\\Frontend\\ContentObject\\ContentObjectRenderer' );
	}

	/**
	 * _initObjectGraphFuncObj( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObjectGraphFuncObj()
	{
		$this->_oGraphFuncObj = GeneralUtility::makeInstance( \TYPO3\CMS\Core\Imaging\GraphicalFunctions::class );
		$this->_oGraphFuncObj->init();
	}

	/**
	 * _initObjectObjectManager( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObjectObjectManager()
	{
		$this->_oObjectManager = GeneralUtility::makeInstance( 'TYPO3\\CMS\\Extbase\\Object\\ObjectManager' );
	}

	/**
	 * _initObjectPlugin( ) : Init localisation
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObjectPlugin()
	{
		$this->_oPlugin = GeneralUtility::makeInstance( \TYPO3\CMS\Frontend\Plugin\AbstractPlugin::class );
	}

	/**
	 * _initObjectTypoScriptService( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initObjectTypoScriptService()
	{
		$this->_oTypoScriptService = GeneralUtility::makeInstance( 'TYPO3\\CMS\\Extbase\\Service\\TypoScriptService' );
	}

	/**
	 * _initVars( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initVars()
	{
		$this->_initVarsFlexform();
		$this->_initVarsCurrent();
		$this->_initVarsArraySearchAndReplace();
//		var_dump( __METHOD__, __LINE__, $this->_aSearchAndReplace );
//		die();
	}

	/**
	 * _initVarsFlexform( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initVarsFlexform()
	{
		$this->_initVarsFlexformPi1();
	}

	/**
	 * _initVarsFlexformPi1( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initVarsFlexformPi1()
	{
		$record = $this->settings[ 'flexform' ][ 'pi1' ];
		$this->_iLaunchId = $record[ 'launchId' ];

		unset( $record[ '_typoScriptNodeValue' ] );
		if( !is_array( $record ) )
		{
			return;
		}

		foreach( ( array ) $record as $key => $value )
		{
			$record[ '%flexform.pi1.' . $key . '%' ] = $value;
			unset( $record[ $key ] );
		}

		$this->_aSearchAndReplace = ( array ) $this->_aSearchAndReplace + ( array ) $record;
	}

	/**
	 * _initVarsArraySearchAndReplace( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initVarsArraySearchAndReplace()
	{
		$records = $this->settings[ 'searchandreplace' ];
		unset( $records[ '_typoScriptNodeValue' ] );

		$records[ 'fileid' ] = '-LaunchId' . $this->_iLaunchId . '-' . $this->_iCurrTime;
		$records[ 'launchid' ] = $this->_iLaunchId;
		$records[ 'time' ] = $this->_iCurrTime;

		if( !is_array( $records ) )
		{
			return;
		}

		foreach( $records as $key => $value )
		{
			$records[ '%' . $key . '%' ] = $value;
			unset( $records[ $key ] );
		}

		$records = $this->_cObjGetSingleArray( $records );
		$this->_aSearchAndReplace = ( array ) $this->_aSearchAndReplace + ( array ) $records;
	}

	/**
	 * _initVars( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _initVarsCurrent()
	{
		$this->_iCurrUid = $GLOBALS[ 'TSFE' ]->id;
		$this->_iCurrPid = $GLOBALS[ 'TSFE' ]->page[ 'pid' ];
		$this->_iCurrTime = time();
	}

	/**
	 * _requirementsContent( ) : 
	 *
	 * @param string $action
	 * @return boolean
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _requirementsContent()
	{
		$directories = $this->settings[ 'configuration' ][ 'directories' ];
		unset( $directories[ '_typoScriptNodeValue' ] );

		$tables = $this->settings[ 'configuration' ][ 'tables' ];
		unset( $tables[ '_typoScriptNodeValue' ] );

		$amount = count( array_keys( $directories ) ) + count( array_keys( $tables ) );

		if( $amount > 0 )
		{
			return true;
		}

		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'true',
								'alertBoxAlertPrompt' => $this->_oPlugin->pi_getLL( 'requirementsContentPromptAlert' ),
								'alertBoxInfoShow' => 'true',
								'alertBoxInfoPrompt' => $this->_oPlugin->pi_getLL( 'requirementsContentPromptInfo' ),
								'alertBoxSuccessShow' => false,
								'alertBoxWarningShow' => 'true',
								'alertBoxWarningPrompt' => $this->_oPlugin->pi_getLL( 'requirementsContentPromptWarning' ),
								'header' => $this->_oPlugin->pi_getLL( 'requirementsInfoHeader' ),
						)
		);
		return false;
	}

	/**
	 * _requirementsExtensions( ) : 
	 *
	 * @param string $action
	 * @return boolean
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _requirementsExtensions()
	{
		$alerts = array();
		$this->_extensions();

		if( empty( $this->_aPrompts ) )
		{
			return true;
		}

		foreach( ( array ) $this->_aPrompts as $key => $prompts )
		{
			$alerts[ $key ][ 'enabled' ] = 'true';
			$alerts[ $key ][ 'prompt' ] = '<ul><li>' . implode( '</li><li>', ( array ) $prompts ) . '</li></ul>';
		}

		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => $alerts[ 'alert' ][ 'enabled' ],
								'alertBoxAlertPrompt' => $alerts[ 'alert' ][ 'prompt' ],
								'alertBoxInfoShow' => $alerts[ 'info' ][ 'enabled' ],
								'alertBoxInfoPrompt' => $alerts[ 'info' ][ 'prompt' ],
								'alertBoxSuccessShow' => false,
								'alertBoxSuccessPrompt' => null,
								'alertBoxWarningShow' => $alerts[ 'warning' ][ 'enabled' ],
								'alertBoxWarningPrompt' => $alerts[ 'warning' ][ 'prompt' ],
								'header' => $this->_oPlugin->pi_getLL( 'requirementsInfoHeader' ),
						)
		);
		return false;
	}

	/**
	 * _requirementsLaunchId( ) : 
	 *
	 * @param string $action
	 * @return boolean
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _requirementsLaunchId()
	{
		if( !empty( $this->_iLaunchId ) )
		{
			return true;
		}
		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'true',
								'alertBoxAlertPrompt' => $this->_oPlugin->pi_getLL( 'requirementsLaunchIdPromptAlert' ),
								'alertBoxInfoShow' => 'true',
								'alertBoxInfoPrompt' => $this->_oPlugin->pi_getLL( 'requirementsLaunchIdPromptInfo' ),
								'header' => $this->_oPlugin->pi_getLL( 'requirementsErrorHeader' ),
						)
		);
		return false;
	}

	/**
	 * _mm( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _mm()
	{
		$pathToXml = $this->settings[ 'configuration' ][ 'tables' ][ $this->_sCurrTable ][ 'xml' ];
		$aXml = $this->_getXmlFileAsArray( $pathToXml );

		foreach( $aXml as $key => $aRecord )
		{
			$aRecord[ 'tx_launch_uid' ] = $this->_sCurrTable . '.' . $key . '.' . $this->_iLaunchId;
			$aRecord = $this->_tableRecordDefaults( $aRecord );
			$aRecord = $this->_tableRecordSearchAndReplace( $aRecord );

			$bPrintOnly = 0;
			$this->_SQLInsert( $this->_sCurrTable, $aRecord, __LINE__, $bPrintOnly );
		}
	}

	/**
	 * _table( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _table()
	{
		$pathToXml = $this->settings[ 'configuration' ][ 'tables' ][ $this->_sCurrTable ][ 'xml' ];
		$aXml = $this->_getXmlFileAsArray( $pathToXml );

		$iterator = new \RecursiveArrayIterator( $aXml );
		iterator_apply( $iterator, 'Netzmacher\Launch\Controller\AbstractController::_tableRecursive', array( $iterator ) );

//		var_dump( __METHOD__, __LINE__ );
//		die();
	}

	/**
	 * _tableRecursive( ) :
	 *
	 * @param object $iterator
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecursive( $iterator )
	{
		while( $iterator->valid() )
		{
			if( !$iterator->hasChildren() )
			{
				$iterator->next();
				continue;
			}

			switch( true )
			{
				case($iterator->key() == 'record'):
					$aRecord = $iterator->current();
					$this->_tableRecord( $aRecord );
					break;
				case($iterator->key() == 'records'):
					$aRecords = $iterator->current();
					$this->_tableRecords( $aRecords );
					break;
				default:
					$this->_sCurrPage = $iterator->key();
					break;
			}

			self::_tableRecursive( $iterator->getChildren() );
			$iterator->next();
		}
	}

	/**
	 * _tableRecord( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecord( $aRecord, $key = null )
	{

		if( $this->_sCurrTable == 'sys_file' )
		{
			$this->_tableSysFileRecord( $aRecord, $key );
			return;
		}

		if( empty( $key ) )
		{
			$key = $this->_sCurrPage;
		}

		$aRecord = $this->_cObjGetSingleArray( $aRecord );
		$aRecord = $this->_tableRecordDefaults( $aRecord );
		$aRecord = $this->_tableRecordSearchAndReplace( $aRecord );
		$aRecord[ 'tx_launch_uid' ] = $key . '.' . $this->_iLaunchId;
		$aRecord[ 'pid' ] = $this->_tableRecordPid( $aRecord, $this->_sCurrPage );

		$isUpdate = $this->settings[ 'configuration' ][ 'tables' ][ $this->_sCurrTable ][ 'isUpdate' ];
		switch( $isUpdate )
		{
			case( true ):
				$this->_tableRecordUpdate( $aRecord );
				break;
			case( false):
			default:
				$this->_tableRecordInsert( $aRecord );
				break;
		}

		$this->_updateVars( $key, $aRecord );
	}

	/**
	 * _tableRecordInsert( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordInsert( $aRecord )
	{
		$bPrintOnly = 0;
		$this->_SQLInsert( $this->_sCurrTable, $aRecord, __LINE__, $bPrintOnly );
	}

	/**
	 * _tableRecordUpdate( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordUpdate( $data )
	{
		//var_dump( __METHOD__, __LINE__, $data, $this->_aSearchAndReplace );
		$field = $this->settings[ 'configuration' ][ 'tables' ][ $this->_sCurrTable ][ 'isUpdate' ];
		$uid = $data[ $field ];
		$where = $field . ' = ' . $uid;
		unset( $data[ $field ] );
		$bPromptQueryOnly = 0;
		$this->_SQLUpdate( $this->_sCurrTable, $where, $data, $uid, __LINE__, $bPromptQueryOnly );
	}

	/**
	 * _tableSysFileRecord( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableSysFileRecord( $aRecord, $key = null )
	{
		$aRecord = $this->_cObjGetSingleArray( $aRecord );
		$aRecord = $this->_tableRecordDefaults( $aRecord );
		$aRecord = $this->_tableRecordSearchAndReplace( $aRecord );

		// Insert record into sys_file
		$file = $this->_fileToDest( $aRecord[ 'file' ] );
		$storageUid = $aRecord[ 'storage' ];
		$uploadfolder = $this->_sysFileStorageUploadfolder( $storageUid );
		$uid = $this->_tableSysFileRecordInsert( $file, $uploadfolder, $storageUid );

		// Update sys_file.record 
		$data[ 'tx_launch_uid' ] = $key . '.' . $this->_iLaunchId;
		$where = 'uid = ' . $uid;
		$bPromptQueryOnly = 0;
		$this->_SQLUpdate( 'sys_file', $where, $data, $uid, __LINE__, $bPromptQueryOnly );

		$this->_updateVars( $key, $aRecord );
	}

	/**
	 * _sysFileStorageUploadfolder( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _sysFileStorageUploadfolder( $storageUid )
	{
		static $aUploadfolder = array();

		if( isset( $aUploadfolder[ $storageUid ] ) )
		{
			return $aUploadfolder[ $storageUid ];
		}

		$fields = 'configuration';
		$table = 'sys_file_storage';
		$where = "uid = " . $storageUid;
		$groupBy = null;
		$orderBy = null;
		$limit = 1;
		$bPromptQueryOnly = 0;

		$res = $this->_SQLSelect( $fields, $table, $where, $groupBy, $orderBy, $limit, __LINE__, $bPromptQueryOnly );
		$row = $GLOBALS[ 'TYPO3_DB' ]->sql_fetch_row( $res );
		$configuration = $row[ 0 ];

		$aXml = GeneralUtility::xml2array( $configuration );
		$sUploadfolder = $aXml[ 'data' ][ 'sDEF' ][ 'lDEF' ][ 'basePath' ][ 'vDEF' ];
		$sUploadfolder = trim( $sUploadfolder, '/' ) . '/';

		$aUploadfolder[ $storageUid ] = $sUploadfolder;

		return $sUploadfolder;
	}

	/**
	 * _tableSysFileRecordInsert( )
	 *
	 * @param	string		$file
	 * @return	integer		Insert uid from sql database
	 * @access private
	 * @internal #i0079
	 * @version 8.0.21
	 * @since   8.0.21
	 */
	private function _tableSysFileRecordInsert( $file, $uploadfolder, $storageUid )
	{
		$pathAbs = PATH_site . $uploadfolder . $file;
		if( !file_exists( $pathAbs ) )
		{
			$header = 'FATAL ERROR: file does not exist';
			$text = ''
							. '<p>'
							. '  The file "' . $pathAbs . '" isn\'t a file or does not exist!<br />'
							. '  Please take care of a proper TypoScript configuration.<br />'
							. '</p>'
			;
			$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
		}

		$identifier = '/' . $file;
		$fileObject = ResourceFactory::getInstance()->getFileObjectByStorageAndIdentifier( $storageUid, $identifier );
		$uid = $fileObject->getUid();
		return $uid;
	}

	/**
	 * _tableRecords( ) :
	 *
	 * @param array $aRecords
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecords( $aRecords )
	{
		foreach( $aRecords as $key => $aRecord )
		{
			$aRecord = $this->_tableRecord( $aRecord, $key );
		}
	}

	/**
	 * _tableRecordDefaults( ) :
	 *
	 * @param array $aRecord
	 * @return array $aRecord
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordDefaults( $aRecord )
	{
		$defaults = $this->settings[ 'configuration' ][ 'tables' ][ $this->_sCurrTable ][ 'defaults' ];
		$defaults = $this->_cObjGetSingleArray( $defaults );

		unset( $defaults[ '_typoScriptNodeValue' ] );

		if( !is_array( $defaults ) )
		{
			return $aRecord;
		}

		$aRecord = ( array ) $aRecord + ( array ) $defaults;
		return $aRecord;
	}

	/**
	 * _tableRecordPid( ) :
	 *
	 * @param string $sCurrPage
	 * @return integer	$pid
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordPid( $aRecord, $sCurrPage )
	{
		if( isset( $aRecord[ 'pid' ] ) )
		{
			return $aRecord[ 'pid' ];
		}

		switch( true )
		{
			case( $this->_sCurrTable == 'pages' ):
				$pid = $this->_tableRecordPidPages( $sCurrPage );
				break;
			case( $this->_sCurrTable != 'pages' ):
			default:
				$pid = $this->_aTableRecordUid[ 'pages' ][ $sCurrPage ];
		}

		if( !empty( $pid ) )
		{
			return $pid;
		}

		var_dump( __METHOD__, __LINE__, 'FATAL ERROR: pid is empty. Current table : "' . $this->_sCurrTable . '".' );
		die();
	}

	/**
	 * _tableRecordPidPages( ) :
	 *
	 * @param string $sCurrPage
	 * @return integer	$pid
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordPidPages( $sCurrPage )
	{
		$pages = explode( '_', $sCurrPage );
		$level = count( $pages );
		switch( true )
		{
			case( $level == 1):
				$pid = $GLOBALS[ 'TSFE' ]->id;
				break;
			case( $level >= 2):
				unset( $pages[ count( $pages ) - 1 ] );
				$parentPage = implode( '_', $pages );
				$pid = $this->_aTableRecordUid[ 'pages' ][ $parentPage ];
				break;
			default:
				var_dump( __METHOD__, __LINE__ . 'FATAL ERROR: level is less then 0. :(' );
				die();
		}

		if( !empty( $pid ) )
		{
			return $pid;
		}

		var_dump( __METHOD__, __LINE__ . 'FATAL ERROR: pid is empty. Current table : ' . $this->_sCurrTable . ', parent page: ' . $parentPage );
		die();
	}

	/**
	 * _tableRecordSearchAndReplace( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tableRecordSearchAndReplace( $aRecord )
	{
		foreach( $aRecord as $key => $value )
		{
			$aRecord[ $key ] = strtr( $value, $this->_aSearchAndReplace );
		}

		return $aRecord;
	}

	/**
	 * _tablesDefault :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tablesDefault( $tables )
	{
		foreach( array_keys( ( array ) $tables ) as $table )
		{
			$this->_sCurrPage = 'root';
			$properties = $this->settings[ 'configuration' ][ 'tables' ][ $table ];
			if( empty( $properties ) )
			{
				continue;
			}
			$isMm = $this->settings[ 'configuration' ][ 'tables' ][ $table ][ 'isMm' ];
			if( !empty( $isMm ) )
			{
				continue;
			}
			$this->_sCurrTable = $table;
			$this->_table();
		}
	}

	/**
	 * _tablesIsMm :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _tablesIsMm( $tables )
	{
		foreach( array_keys( ( array ) $tables ) as $table )
		{
			$properties = $this->settings[ 'configuration' ][ 'tables' ][ $table ];
			if( empty( $properties ) )
			{
				continue;
			}
			$isMm = $this->settings[ 'configuration' ][ 'tables' ][ $table ][ 'isMm' ];
			if( empty( $isMm ) )
			{
				continue;
			}
			$this->_sCurrTable = $table;
			$this->_mm();
		}
	}

	/**
	 * _updateVars( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _updateVars( $key, $aRecord )
	{
		if( empty( $key ) )
		{
			$key = $this->_sCurrPage;
		}

		$uid = $this->_aTableSQLLastID[ $this->_sCurrTable ];
		$markerUid = '%' . $this->_sCurrTable . ':' . $key . '.uid%';

		$label = $this->_updateVarsLabel( $aRecord );
		$markerLabel = '%' . $this->_sCurrTable . ':' . $key . '.title%';

		if( !isset( $this->_aTableRecordUid[ $this->_sCurrTable ][ $key ] ) )
		{
			$this->_aTableRecordUid[ $this->_sCurrTable ][ $key ] = $uid;
			$this->_aSearchAndReplace[ $markerUid ] = $uid;
			$this->_aSearchAndReplace[ $markerLabel ] = $label;
			return;
		}

		$header = 'FATAL ERROR: non unique uid';
		$text = ''
						. '<p>'
						. '  You have used a tag in your XML-files twice.<br />'
						. '  This can affect unproper relations.<br />'
						. '  Workflow of migration is canceled.<br />'
						. '  Please take care of XML-files with unique tags.<br />'
						. '  <br />'
						. '  Here: ' . $this->_sCurrTable . '. Non unique tag: ' . $key . '.<br />'
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, __METHOD__, __LINE__ );
	}

	/**
	 * _updateVarsLabel( ) :
	 *
	 * @param array $aRecord
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _updateVarsLabel( $aRecord )
	{
		if( isset( $aRecord[ 'title' ] ) )
		{
			return $aRecord[ 'title' ];
		}
		if( isset( $aRecord[ 'header' ] ) )
		{
			return $aRecord[ 'header' ];
		}
		if( isset( $aRecord[ 'label' ] ) )
		{
			return $aRecord[ 'label' ];
		}
		if( isset( $aRecord[ 'label' ] ) )
		{
			return $aRecord[ 'label' ];
		}
		if( isset( $aRecord[ 'file' ] ) )
		{
			return $aRecord[ 'file' ];
		}

		return 'noTitle';
	}

	/**
	 * _zzDieWiPrompt( ) :
	 *
	 * @access private
	 * @return boolean
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _zzDieWiPrompt( $header, $text, $method, $line )
	{
		$prompt = '
      <h1 style="color:red;">
        ' . $header . '
      </h1>
      ' . $text . '
      <p>
        Sorry for the trouble. Launch - TYPO3-Installer<br />
        Error occures here: ' . $method . ' at #' . $line . '
      </p>
      <p>
        If you need any help, please visit the
        <a href="https://typo3-browser-forum.de/foren/launch/" target="_blank" title="Forum Launch - TYPO3-Installer. 500 TYPO3-integrators are registered.">
          Forum Launch &raquo;</a>
      </p>
      ';
		die( $prompt );
	}

	/**
	 * _zzDieWoFile( ) :
	 *
	 * @param string  $filename
	 * @param integer $line
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _zzDieWoFile( $filename, $method, $line )
	{
		if( is_file( $filename ) )
		{
			return;
		}

		$header = 'FATAL ERROR';
		$text = ''
						. '<p>'
						. '  The file ' . $filename . ' isn\'t a file or does not exist!<br />'
						. '  Please take care of a proper TypoScript configuration.<br />'
						. '</p>'
		;
		$this->_zzDieWiPrompt( $header, $text, $method, $line );
	}

	/**
	 * commit( ) :
	 *
	 * @return void
	 * @access protected
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	protected function commit()
	{
		$this->_directories();

		$tables = $this->settings[ 'configuration' ][ 'tables' ];
		unset( $tables[ '_typoScriptNodeValue' ] );
		$this->_tablesDefault( $tables );
		$this->_tablesIsMm( $tables );

		$this->view->assignMultiple(
						array(
								'pageUid' => $this->_getUidLaunch(),
						)
		);

		$searchAndReplace = $this->settings[ 'development' ][ 'prompts' ][ 'searchAndReplace' ];
		if( !$searchAndReplace )
		{
			return;
		}

		$devPrompt = '<pre>' . var_export( $this->_aSearchAndReplace, true ) . '</pre>';

		$this->view->assignMultiple(
						array(
								'devPrompt' => $devPrompt,
						)
		);
	}

	/**
	 * delete( ) :
	 *
	 * @return void
	 * @access protected
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	protected function delete()
	{
		$this->_deleteDirectories( );
		$this->_deleteTables();
	}

	/**
	 * init( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	protected function init()
	{
		$this->_initLL();
		$this->_initExtConf();
		$this->_initObject();
		$this->_initVars();
	}

	/**
	 * requirements( ) : 
	 *
	 * @param string $action
	 * @return boolean
	 * @access protected
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	protected function requirements()
	{
		if( !$this->_requirementsContent() )
		{
			return false;
		}

		if( !$this->_requirementsLaunchId() )
		{
			return false;
		}

		if( !$this->_requirementsExtensions() )
		{
			return false;
		}

		return true;
	}

}
