<?php

namespace Netzmacher\Launch\Controller;

/* * *************************************************************
 *  Copyright notice
 *
 *  (c) 2017 - Dirk Wildt <http://wildt.at.die-netzmacher.de>
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 * ************************************************************* */

/**
 * Class for frontend editing
 *
 * @package TYPO3
 * @subpackage launch
 * @author Dirk Wildt <http://wildt.at.die-netzmacher.de>
 * @version 2.0.0
 * @since 2.0.0
 */
class LaunchController extends AbstractController
{

	/**
	 * _viewAssignCommit( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _viewAssignCommit()
	{
		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'false',
								'alertBoxInfoShow' => 'false',
								'alertBoxSuccessShow' => 'true',
								'alertBoxSuccessPrompt' => $this->_oPlugin->pi_getLL( 'viewCommitDefaultPrompt' ),
								'alertBoxWarningShow' => 'false',
								'header' => $this->_oPlugin->pi_getLL( 'viewCommitDefaultHeader' ),
								'buttonValueReload' => $this->_oPlugin->pi_getLL( 'buttonValueReload' ),
								'buttonValueShow' => $this->_oPlugin->pi_getLL( 'buttonValueShow' ),
						)
		);
	}

	/**
	 * _viewAssignDelete( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _viewAssignDelete()
	{
		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'false',
								'alertBoxInfoShow' => 'false',
								'alertBoxWarningShow' => 'false',
								'header' => $this->_oPlugin->pi_getLL( 'viewDeleteDefaultHeader' ),
								'prompt' => $this->_oPlugin->pi_getLL( 'viewDeleteDefaultPrompt' ),
								'buttonValueReload' => $this->_oPlugin->pi_getLL( 'buttonValueReload' ),
								'buttonValueDelete' => $this->_oPlugin->pi_getLL( 'buttonValueDelete' ),
						)
		);
	}

	/**
	 * _viewAssignLaunch( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _viewAssignLaunch()
	{
		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'false',
								'alertBoxInfoShow' => 'false',
								'alertBoxWarningShow' => 'false',
								'header' => $this->_oPlugin->pi_getLL( 'viewLaunchDefaultHeader' ),
								'prompt' => $this->_oPlugin->pi_getLL( 'viewLaunchDefaultPrompt' ),
								'buttonValueCommit' => $this->_oPlugin->pi_getLL( 'buttonValueCommit' ),
								'buttonValueShow' => $this->_oPlugin->pi_getLL( 'buttonValueShow' ),
								'buttonValueUndo' => $this->_oPlugin->pi_getLL( 'buttonValueUndo' ),
						)
		);
		
		$pageUid = $this->_getUidLaunch();
		if( empty( $pageUid ) )
		{
			return;
		}
		
		$this->view->assignMultiple(
						array(
								'pageUid' => $pageUid,
						)
		);
	}

	/**
	 * _viewAssignUndo( ) :
	 *
	 * @return void
	 * @access private
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	private function _viewAssignUndo()
	{
		$this->view->assignMultiple(
						array(
								'alertBoxAlertShow' => 'false',
								'alertBoxInfoShow' => 'false',
								'alertBoxWarningShow' => 'false',
								'header' => $this->_oPlugin->pi_getLL( 'viewUndoDefaultHeader' ),
								'prompt' => $this->_oPlugin->pi_getLL( 'viewUndoDefaultPrompt' ),
								'buttonValueReload' => $this->_oPlugin->pi_getLL( 'buttonValueReload' ),
								'buttonValueDelete' => $this->_oPlugin->pi_getLL( 'buttonValueDelete' ),
						)
		);
	}

	/**
	 * commitAction( ) : 
	 *
	 * @return void
	 * @access public
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	public function commitAction()
	{
		$this->init();
		$this->_viewAssignCommit();
		if( !$this->requirements() )
		{
			return;
		}
		$this->commit();
	}

	/**
	 * deleteAction( ) : 
	 *
	 * @return void
	 * @access public
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	public function deleteAction()
	{
		$this->init();
		$this->_viewAssignDelete();
		$this->delete();
	}

	/**
	 * launchAction( ) : 
	 *
	 * @return void
	 * @access public
	 * @version 2.0.0
	 * @since 2.0.0
	 *
	 */
	public function launchAction()
	{
		$this->init();
		$this->_viewAssignLaunch();
	}

	/**
	 * undoAction( ) : 
	 *
	 * @return void
	 * @access public
	 * @version 2.0.0
	 * @since 2.0.0
	 */
	public function undoAction()
	{
		$this->init();
		$this->_viewAssignUndo();
	}

}
